"""
Bodge: Python package for efficient tight-binding modeling of superconductors.

For more information, please consult the bundled documentation and examples.
"""

from .common import *
from .hamiltonian import *
from .lattice import *

__author__ = "Jabir Ali Ouassou"
__version__ = "1.1.1"
__all__ = [
    # Core library
    "Lattice",
    "CubicLattice",
    "Hamiltonian",
    "Coord",
    "Coords",
    "Index",
    "Indices",
    # Helper functions
    "ssd",
    "swave",
    "pwave",
    "dwave",
    # Useful constants
    "π",
    "σ",
    "σ0",
    "σ1",
    "σ2",
    "σ3",
    "jσ",
    "jσ0",
    "jσ1",
    "jσ2",
    "jσ3",
    # ASCII alternatives
    "pi",
    "sigma",
    "sigma0",
    "sigma1",
    "sigma2",
    "sigma3",
    "jsigma",
    "jsigma0",
    "jsigma1",
    "jsigma2",
    "jsigma3",
]



#!/usr/bin/env python

"""Benchmarking of Bodge vs. Kwant for constructing a Bogoliubov-deGennes Hamiltonian.

Note that this requires a `pip install kwant` to get the library benchmarked against.
"""

from time import sleep

import kwant
import numpy as np

from bodge import *

t = 1  # Hopping parameter
μ = -3 * t  # Chemical potential
M0 = 1.5 * t  # Magnetic exchange field
Δ0 = 0.1 * t  # Superconducting gap
χ = 0.5  # Superconducting phase winding


def timer(timers=[]):
    """Simple timer used for benchmarking."""
    from time import time

    # Save the current timestamp.
    timers.append(time())

    # Difference since last time stamp.
    try:
        return timers[-1] - timers[-2]
    except:
        return None


def bench_kwant(L, W, sparse=True):
    """Construct a Bogoliubov-deGennes Hamiltonian on a square lattice using Kwant."""
    # Define Pauli matrices for spin degree of freedom.
    σ_0 = np.array([[1, 0], [0, 1]])
    σ_1 = np.array([[0, 1], [1, 0]])
    σ_2 = np.array([[0, -1j], [1j, 0]])
    σ_3 = np.array([[1, 0], [0, -1]])

    # Define Pauli matrices for particle-hole degree of freedom.
    τ_0 = σ_0.copy()
    τ_1 = σ_1.copy()
    τ_2 = σ_2.copy()
    τ_3 = σ_3.copy()

    # Sleep before benchmarking.
    sleep(0.01)

    # Start a timer.
    timer()

    # Construct the lattice and Hamiltonian.
    lattice = kwant.lattice.square(norbs=4)
    system = kwant.Builder()

    # On-site terms in the Hamiltonian.
    for x in range(L):
        φ = χ * x / L
        τ_φ = 1j * (τ_2 * np.cos(φ) + τ_1 * np.sin(φ))

        for y in range(W):
            if x < L // 2:
                # Superconducting region.
                system[lattice(x, y)] = -μ * np.kron(τ_3, σ_0) - Δ0 * np.kron(τ_φ, 1j * σ_2)
            else:
                # Ferromagnetic region.
                system[lattice(x, y)] = -μ * np.kron(τ_3, σ_0) - M0 * np.kron(τ_3, σ_3)

    # Hopping terms in the Hamiltonian.
    for x in range(L - 1):
        for y in range(W):
            system[lattice(x, y), lattice(x + 1, y)] = -t * np.kron(τ_3, σ_0)
    for x in range(L):
        for y in range(W - 1):
            system[lattice(x, y), lattice(x, y + 1)] = -2 * t * np.kron(τ_3, σ_0)

    # Compile the Hamiltonian object to a matrix.
    system = system.finalized()
    H = system.hamiltonian_submatrix(sparse=sparse)

    # Benchmark results.
    T = timer()

    return H, T


def bench_bodge(L, W, sparse=True):
    """Construct a Bogoliubov-deGennes Hamiltonian on a square lattice using Bodge."""
    # Sleep before benchmarking.
    sleep(0.01)

    # Start a timer.
    timer()

    # Construct the lattice and Hamiltonian.
    lattice = CubicLattice((L, W, 1))
    system = Hamiltonian(lattice)

    # Fill out the Hamiltonian.
    with system as (H, Δ):
        # On-site terms.
        for i in lattice.sites():
            if i[0] < L // 2:
                # Superconducting region.
                H[i, i] = -μ * σ0
                Δ[i, i] = -Δ0 * np.exp(1j * χ * i[0] / L) * jσ2
            else:
                # Ferromagnetic region.
                H[i, i] = -μ * σ0 - M0 * σ3

        # Hopping terms.
        for i, j in lattice.bonds(axis=0):
            H[i, j] = -t * σ0
        for i, j in lattice.bonds(axis=1):
            H[i, j] = -2 * t * σ0

    # Compile the Hamiltonian object to a matrix.
    if sparse:
        H = system.matrix(format="csr")
    else:
        H = system.matrix(format="dense")

    # Benchmark results.
    T = timer()

    return H, T


if __name__ == "__main__":
    for L in [2**n for n in range(1, 12)]:
        for W in [L, 2 * L]:
            # Number of iterations. We repeat brief calculations to get better statistics.
            if L * W < 1e4:
                N = 30
            else:
                N = 1

            # Benchmark the matrix construction.
            T1s = []
            T2s = []
            for i in range(N):
                H1, T1 = bench_bodge(L, W)
                H2, T2 = bench_kwant(L, W)

                T1s.append(T1)
                T2s.append(T2)

            T1 = np.min(T1s)
            T2 = np.min(T2s)

            # Report the benchmarks.
            print(f"{L*W}\tBodge\t{T1}")
            print(f"{L*W}\tKwant\t{T2}")

            # Ensure the results are equal.
            print(f"{L*W}\tDiff\t{np.max(np.abs(H1 - H2))}")



from .common import *


class Lattice:
    """Base class that represents a general atomic lattice (in 1D, 2D, or 3D).

    This abstract class defines a general interface for iterating over all
    sites (atoms) and bonds (nearest neighbors) in a lattice. In the language
    of graph theory, this class lets us traverse all the nodes and links of a
    simple graph. The actual graph traversal must be defined by subclassing
    `Lattice` and implementing all methods that raise `NotImplementedError`.

    Note that you are free to implement optional arguments to these methods.
    For instance, it may be useful to iterate over one sublattice at a time
    when calling `.sites` on a honeycomb lattice, or to iterate over the x-
    and y-axes separately when calling `.bonds` on a rectangular lattice.
    However, it must be possible to call both methods without additional
    arguments to traverse all sites and bonds in the lattice, respectively.
    This is assumed by the class `Hamiltonian` which is built on `Lattice`.
    """

    @typecheck
    def __init__(self, shape: Coord):
        # Create an abstract class.
        if self.__class__.__name__ == "Lattice":
            raise ValueError("This class is not intended to be instantiated directly.")

        # Number of atoms per dimension.
        self.shape: Coord = shape

        # Number of atoms in the lattice.
        self.size: Index = int(np.prod(shape))

        # Dimensionality of the lattice.
        self.dim: int = sum([1 for x in self.shape if x > 1])

    @typecheck
    def __getitem__(self, coord: Coord) -> Index:
        """Syntactic sugar for converting coordinates into indices."""
        return self.index(coord)

    @typecheck
    def __iter__(self) -> Iterator[Coords]:
        """Iterate over all on-site and nearest-neighbor interactions."""
        for index in self.sites():
            yield (index, index)
        for indices in self.bonds():
            yield indices
        for indices in self.edges():
            yield indices

    @typecheck
    def __repr__(self) -> str:
        """Representation of the object for `print()`."""
        return self.__class__.__name__ + str(self.shape)

    @typecheck
    def index(self, coord: Coord) -> Index:
        """Convert a 3D site coordinate to a 1D index."""
        raise NotImplementedError

    @typecheck
    def sites(self) -> Iterator[Coord]:
        """Iterate over all atomic sites in the lattice."""
        raise NotImplementedError

    @typecheck
    def bonds(self) -> Iterator[Coords]:
        """Iterate over all atomic bonds in the lattice.

        The intended usage is that `for i, j in lattice.bonds()` should
        yield all nearest-neighbor sites (i, j) in the lattice.
        """
        raise NotImplementedError

    @typecheck
    def edges(self) -> Iterator[Coords]:
        """Iterate over pairs of edges in the lattice.

        The intended usage is that `for i, j in lattice.edges()` should yield
        sites i and j on *opposite* edges of a system. Thus, hopping terms
        between such sites should result in periodic boundary conditions.
        """
        raise NotImplementedError


class CubicLattice(Lattice):
    """Concrete representation of a primitive cubic lattice.

    The constructor takes as its argument the number of lattice sites along each
    direction in a 3D Euclidean space. So a 10x10x10 cubic lattice is given by:

    >>> lattice = CubicLattice((10, 10, 10))

    The same class can be used to model square lattices or rectangular lattices
    as well. To e.g. construct an 30x30 square lattice, simply do the following:

    >>> lattice = CubicLattice((30, 30, 1))
    """

    @typecheck
    def index(self, coord: Coord) -> Index:
        """Convert a 3D site coordinate to a 1D index."""
        for i in range(3):
            if coord[i] < 0 or coord[i] >= self.shape[i]:
                raise ValueError(f"Coordinate {coord} out of bounds")
        else:
            return coord[2] + coord[1] * self.shape[2] + coord[0] * self.shape[1] * self.shape[2]

    @typecheck
    def sites(self) -> Iterator[Coord]:
        """Iterate over all atomic sites in the lattice.

        Thus, e.g. `for i in lattice.sites()` iterates over all lattice sites.
        """
        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                for z in range(self.shape[2]):
                    yield (x, y, z)

    @typecheck
    def bonds(self, axis: int | None = None) -> Iterator[Coords]:
        """Iterate over all atomic bonds in the lattice.

        The `axis` argument allows iterating over bonds along only one
        cardinal axis. For example, `for i, j in lattice.bonds(axis=0)`
        would iterate over nearest neighbors connected along the x-axis.

        If `axis` is not set, we iterate over bonds in all directions.
        """
        match axis:
            case None:
                # Neighbors along all axes.
                yield from self.bonds(axis=2)
                yield from self.bonds(axis=1)
                yield from self.bonds(axis=0)
            case 0:
                # Neighbors along x-axis.
                for x in range(self.shape[0] - 1):
                    for y in range(self.shape[1]):
                        for z in range(self.shape[2]):
                            yield (x, y, z), (x + 1, y, z)
                            yield (x + 1, y, z), (x, y, z)
            case 1:
                # Neighbors along y-axis.
                for x in range(self.shape[0]):
                    for y in range(self.shape[1] - 1):
                        for z in range(self.shape[2]):
                            yield (x, y, z), (x, y + 1, z)
                            yield (x, y + 1, z), (x, y, z)
            case 2:
                # Neighbors along z-axis.
                for x in range(self.shape[0]):
                    for y in range(self.shape[1]):
                        for z in range(self.shape[2] - 1):
                            yield (x, y, z), (x, y, z + 1)
                            yield (x, y, z + 1), (x, y, z)
            case _:
                raise ValueError("No such axis")

    @typecheck
    def edges(self, axis: int | None = None) -> Iterator[Coords]:
        """Iterate over pairs of atoms at opposite lattice edges.

        For instance, if you want periodic boundary conditions along the
        x-axis, you can use `for i, j in lattice.edges(axis=0)` to find the
        lattice sites (i, j) that you should connect via hopping terms.

        If `axis` is not set, we iterate over edges in all directions.
        """
        Lx, Ly, Lz = self.shape[0], self.shape[1], self.shape[2]
        match axis:
            case None:
                # Edges along all axes.
                yield from self.edges(axis=2)
                yield from self.edges(axis=1)
                yield from self.edges(axis=0)
            case 0:
                # Edges at x=0 and x=Lx-1.
                for y in range(Ly):
                    for z in range(Lz):
                        yield (0, y, z), (Lx - 1, y, z)
                        yield (Lx - 1, y, z), (0, y, z)
            case 1:
                # Edges at y=0 and y=Ly-1.
                for x in range(Lx):
                    for z in range(Lz):
                        yield (x, 0, z), (x, Ly - 1, z)
                        yield (x, Ly - 1, z), (x, 0, z)
            case 2:
                # Edges at z=0 and z=Lz-1.
                for x in range(Lx):
                    for y in range(Ly):
                        yield (x, y, 0), (x, y, Lz - 1)
                        yield (x, y, Lz - 1), (x, y, 0)
            case _:
                raise ValueError("No such axis")



import pytest

from bodge.common import *


def test_pauli():
    # Test that the quaternion identities hold.
    assert np.allclose(σ1 @ σ1, σ0)
    assert np.allclose(σ2 @ σ2, σ0)
    assert np.allclose(σ3 @ σ3, σ0)

    assert np.allclose(σ1 @ σ2, jσ3)
    assert np.allclose(σ2 @ σ3, jσ1)
    assert np.allclose(σ3 @ σ1, jσ2)

    assert np.allclose(σ1 @ σ2 @ σ3, jσ0)



from pytest import raises

from bodge.common import *
from bodge.lattice import *


def test_abc():
    # Abstract base class should not be constructable.
    with raises(ValueError):
        lat = Lattice((1, 1, 1))

    # Derived classes are constructable but must override other methods to work.
    class MyLattice(Lattice):
        pass

    lat = MyLattice((1, 2, 3))
    with raises(NotImplementedError):
        lat[(0, 0, 0)]
    with raises(NotImplementedError):
        lat.sites()
    with raises(NotImplementedError):
        lat.bonds()
    with raises(NotImplementedError):
        lat.edges()

    # Derived classes should have a readable format.
    assert str(lat) == "MyLattice(1, 2, 3)"


def test_cubic_sites():
    lat = CubicLattice((3, 5, 7))
    tot = 0
    for ind, site in enumerate(lat.sites()):
        tot += 1

        # Verify that indexing is contiguous.
        assert ind == lat[site]

        # Verify that index bounds are satisfied.
        assert site[0] >= 0 and site[0] < lat.shape[0]
        assert site[1] >= 0 and site[1] < lat.shape[1]
        assert site[2] >= 0 and site[2] < lat.shape[2]

    # Verify that number of elements is correct.
    assert tot == 3 * 5 * 7

    # Verify that we get out-of-bounds errors.
    with raises(ValueError):
        lat[(-1, 0, 0)]
    with raises(ValueError):
        lat[(0, -1, 0)]
    with raises(ValueError):
        lat[(0, 0, -1)]
    with raises(ValueError):
        lat[(3, 0, 0)]
    with raises(ValueError):
        lat[(0, 5, 0)]
    with raises(ValueError):
        lat[(0, 0, 7)]


def test_cubic_bonds():
    tot = 0
    lat = CubicLattice((2, 3, 5))

    # Verify neighbor coordinates along each axis.
    for (x1, y1, z1), (x2, y2, z2) in lat.bonds(axis=0):
        tot += 1
        assert (x2 == x1 + 1 or x2 == x1 - 1) and y2 == y1 and z2 == z1
    for (x1, y1, z1), (x2, y2, z2) in lat.bonds(axis=1):
        tot += 1
        assert x2 == x1 and (y2 == y1 + 1 or y2 == y1 - 1) and z2 == z1
    for (x1, y1, z1), (x2, y2, z2) in lat.bonds(axis=2):
        tot += 1
        assert x2 == x1 and y2 == y1 and (z2 == z1 + 1 or z2 == z1 - 1)

    # Verify the total number of nearest-neighbor pairings. Note the prefactor 2,
    # since "forwards" and "backwards" pairings are both counted here.
    assert tot == 2 * ((2 - 1) * 3 * 5 + 2 * (3 - 1) * 5 + 2 * 3 * (5 - 1))

    # Verify that the bounds checks works.
    with raises(ValueError):
        for i, j in lat.bonds(axis=3):
            print(i, j)


def test_cubic_edges():
    tot = 0
    lat = CubicLattice((2, 3, 5))

    # Verify that each coordinate is along an edge.
    for (x1, y1, z1), (x2, y2, z2) in lat.edges(axis=0):
        tot += 1
        assert x1 == 0 or x2 == 0
        assert x1 == 1 or x2 == 1
    for (x1, y1, z1), (x2, y2, z2) in lat.edges(axis=1):
        tot += 1
        assert y1 == 0 or y2 == 0
        assert y1 == 2 or y2 == 2
    for (x1, y1, z1), (x2, y2, z2) in lat.edges(axis=2):
        tot += 1
        assert z1 == 0 or z2 == 0
        assert z1 == 4 or z2 == 4

    # Verify that the number of edge sites is correct.
    assert tot == 2 * ((2 * 3) + (3 * 5) + (5 * 2))

    # Verify that the bounds checks works.
    with raises(ValueError):
        for i, j in lat.edges(axis=3):
            print(i, j)



from .common import *
from .lattice import Lattice


class Hamiltonian:
    """Tight-binding representation of a physical system.

    This class can be used to construct Hamiltonian matrices for condensed
    matter systems with particle-hole and spin degrees of freedom. Instead of
    explicitly constructing the whole matrix, this class allows you to specify
    the minimum number of matrix elements required via a `with` block, and the
    the remainder is autofilled via symmetries. Moreover, it allows you to use
    lattice coordinates instead of matrix indices to fill out these elements.

    Internally, this constructs a block-sparse matrix (BSR format), enabling
    the construction of megadimensional tight-binding systems (>10^6 sites).
    However, you can use the `.matrix()` method to export the result to other
    common sparse matrix formats (e.g. CSR) or a dense matrix (NumPy array).
    You can also use the `.diagonalize()` method to get the eigenvalues and
    eigenvectors of the Hamiltonian, although this is usually inefficient.

    For examples of how to use this class, see the bundled documentation.
    """

    @typecheck
    def __init__(self, lattice: Lattice):
        # Lattice instance used as basis coordinates for the system.
        self.lattice: Lattice = lattice

        # Number of lattice points that the system has. The integers in front of
        # the lattice size are the local degrees of freedom at each lattice site.
        self.shape: Indices = (4 * lattice.size, 4 * lattice.size)

        # Initialize the most general 4N×4N Hamiltonian for this lattice as a
        # sparse matrix. The COO format is most efficient for constructing the
        # matrix, but the BSR format is more computationally efficient later.
        sites = sum(1 for _ in lattice.sites())
        bonds = sum(2 for _ in lattice.bonds())
        edges = sum(2 for _ in lattice.edges())
        size = sites + bonds + edges

        rows = np.zeros(size, dtype=np.int64)
        cols = np.zeros(size, dtype=np.int64)
        data = np.repeat(np.int8(1), size)

        k = 0
        for ri, rj in lattice:
            i, j = 4 * lattice[ri], 4 * lattice[rj]

            rows[k] = i
            cols[k] = j
            k += 1

            if i != j:
                rows[k] = j
                cols[k] = i
                k += 1

        skeleton = CooMatrix((data, (rows, cols)), shape=self.shape).tobsr((4, 4))

        # Save a complex matrix that encodes the Hamiltonian matrix itself.
        # Each element is set to zero and must later be populated for use.
        self._matrix: BsrMatrix = BsrMatrix(skeleton, dtype=np.complex128)
        self._matrix.data[...] = 0

        # Simplify direct access to the underlying data structure.
        self._data: Matrix = self._matrix.data

    @typecheck
    def __enter__(self) -> tuple[dict[Coords, Matrix], dict[Coords, Matrix]]:
        """Implement a context manager interface for the class.

        This lets us write compact `with` blocks like the below, which is much
        more convenient than having to construct the matrix elements explicitly.

        ```python
        with system as (H, Δ):
            H[i, j] = ...
            Δ[i, j] = ...
        ```

        Note that the `__exit__` method is responsible for actually transferring
        all the elements of H and Δ to the correct locations in the Hamiltonian.
        """
        # Prepare storage for the context manager.
        self._hopp = {}
        self._pair = {}

        return self._hopp, self._pair

    @typecheck
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Implement a context manager interface for the class.

        This part of the implementation takes care of finalizing the Hamiltonian:

        - Transferring elements from the context manager dicts to the sparse matrix;
        - Ensuring that particle-hole and nearest-neighbor symmetries are respected;
        - Verifying that the constructed Hamiltonian is actually Hermitian.
        """
        # Process hopping terms: H[i, j].
        for (i, j), val in self._hopp.items():
            # Find this matrix block.
            k = self.index(i, j)

            # Update respecting electron-hole symmetry.
            self._data[k, 0:2, 0:2] = +val
            self._data[k, 2:4, 2:4] = -val.conj()

        # Process pairing terms: Δ[i, j].
        for (i, j), val in self._pair.items():
            # Find this matrix block.
            k1 = self.index(i, j)
            k2 = self.index(j, i)

            # Update respecting Hermitian symmetry.
            self._data[k1, 0:2, 2:4] = +val
            self._data[k2, 2:4, 0:2] = +val.T.conj()

        # Verify that the matrix is Hermitian.
        if np.max(np.abs(self._matrix - self._matrix.getH())) > 1e-6:
            raise RuntimeError("The constructed Hamiltonian is not Hermitian!")

        # Reset accessors.
        del self._hopp
        del self._pair

    @typecheck
    def matrix(self, format: str = "dense") -> SpMatrix | Matrix:
        """Return a specific matrix representation of the Hamiltonian.

        If the format is set to "dense", the result is returned as a NumPy
        array. If it is set to one of {"bsr", "csr", "csc"}, then the
        corresponding SciPy sparse matrix format is used instead.
        """
        # Transform the stored matrix into the requested matrix format. Trim
        # any remaining zero entries if a sparse matrix is requested.
        match format:
            case "bsr":
                # NOTE: Don't run eliminate_zeros on the original matrix, as that
                # would preclude adding new elements to the Hamiltonian later.
                H: BsrMatrix = self._matrix.copy()
                H.eliminate_zeros()
            case "csr":
                H: CsrMatrix = self._matrix.tocsr()
                H.eliminate_zeros()
            case "csc":
                H: CscMatrix = self._matrix.tocsc()
                H.eliminate_zeros()
            case "dense":
                H: Matrix = self._matrix.todense()
            case _:
                raise RuntimeError("Requested matrix format is not yet supported")

        return H

    @typecheck
    def index(self, row: Coord, col: Coord) -> Index:
        """Determine the sparse matrix index corresponding to block (row, col).

        This can be used to access `self.data[index, :, :]` when direct
        changes to the encapsulated block-sparse matrix are required.
        """
        indices, indptr = self._matrix.indices, self._matrix.indptr

        i, j = self.lattice[row], self.lattice[col]
        js = indices[indptr[i] : indptr[i + 1]]
        k = indptr[i] + np.where(js == j)

        return Index(k[0, 0])

    @typecheck
    def diagonalize(self) -> tuple[Matrix, Matrix]:
        """Calculate the exact eigenstates of the system via direct diagonalization.

        This calculates and returns the eigenvalues and eigenvectors of the system.
        If you run it as `E, v = system.diagonalize()`, then the eigenvalue `E[n]`
        corresponds to the eigenvector `v[n, :, :]`, where the remaining indices
        of the vector correspond to a position index and a Nambu⊗Spin index. The
        last one means that at a given lattice site `i`, `v[n, i, 0:3]` will give
        you the eigenvector components corresponding to indices {e↑, e↓, h↑, h↓}.

        Please note that this function uses *dense* matrices and thus requires
        a lot of compute power and memory for large matrices. Many computations
        can be performed using sparse matrix algorithms instead, so if you have
        a lattice with millions of sites I would consider alternative methods.
        """
        # Convert to a dense matrix.
        H = self.matrix(format="dense")

        # Calculate the relevant eigenvalues and eigenvectors.
        eigval, eigvec = la.eigh(H, subset_by_value=(0.0, np.inf), overwrite_a=True, driver="evr")
        eigval = np.array(eigval)
        eigvec = np.array(eigvec)

        # Restructure the eigenvectors to have the format eigvec[n, i, α],
        # where n corresponds to eigenvalue E[n], i is a position index, and
        # α represents the combined particle and spin index {e↑, e↓, h↑, h↓}.
        eigvec = eigvec.T.reshape((eigval.size, -1, 4))

        return eigval, eigvec

    @typecheck
    def free_energy(self, temperature: float = 0.0) -> float:
        """Calculate the Landau free energy for a given Hamiltonian.

        This is done by computing all the positive eigenvalues ε_n of the matrix,
        and subsequently evaluating the entropy contributions to the free energy.
        The resulting free energy is then formulated as F = U - TS, where U is
        the internal energy (calculated from ɛ_n), T is the provided temperature,
        and S is the system's entropy (calculated from ε_n and T).

        In general, U has a constant contribution as well which corresponds to
        the non-matrix parts to the Hamiltonian operator. This contribution
        can't be calculated by this method as it is not part of the Hamiltonian
        matrix. For non-selfconsistent calculations, you usually don't need this
        constant. However, since this "constant" contribution is generally a
        function of the mean fields, you need to add it to the return value of
        this function for selfconsistent calculations to be correct.

        The algorithm implemented here is explained in Appendix C of:

            Ouassou et al. PRB 109, 174506 (2024).
            DOI: 10.1103/PhysRevB.109.174506
        """
        T = temperature
        H = self.matrix(format="dense")

        # Calculate the eigenvalues via a dense parallel algorithm. My benchmarks
        # have shown that this is usually faster than using the sparse solver.
        ε = la.eigh(H, overwrite_a=True, eigvals_only=True, driver="evr")

        # Extract the positive eigenvalues.
        ε = ε[ε > 0]

        # Internal energy.
        U = -(1 / 2) * np.sum(ε)

        # Entropy contribution.
        if T == 0:
            S = 0
        elif T > 0:
            S = np.sum(np.log(1 + np.exp(-ε / T)))
        else:
            raise ValueError("Expected non-negative temperature!")

        # Free energy
        F = U - T * S

        return F

    def ldos(self, site: Coord, energies: Matrix) -> tuple[Matrix, Matrix]:
        """Calculate the local density of states via a resolvent operator approach.

        We define the resolvent operator as [(ε+iη)I - H] R(ε) = I, which can be
        divided into vectors R(ε) = [r_1 ... r_N] and I = [e_1 ... e_N]. Diagonal
        elements of the resolvent matrix correspond to the density of states. By
        calculating one vector at a time via a sparse linear solver `spsolve`, the
        local density of states can be efficiently calculated at specific points.

        Note that if the requested energies are symmetric around ε = 0, then only
        half the calculations need to be performed due to particle-hole symmetry.

        The algorithm implemented here is explained in detail in Appendix A of:

        Ouassou et al. PRB 109, 174506 (2024).
        DOI: 10.1103/PhysRevB.109.174506
        """
        # Sparse Hamiltonian and identity matrices.
        H = self.matrix(format="csc")
        I = sp.identity(H.shape[0], format="csc")

        # Ensure that the input is NumPy floats.
        energies = np.array(energies, dtype=float)

        # Determine which energies we need calculations for.
        ε = np.unique(np.abs(energies))

        # Determine broadening parameter from the provided energies.
        Γ = np.gradient(ε)

        # Construct a reduced identity matrix for the relevant coordinate.
        N = H.shape[1]
        M = 4

        i = self.lattice[site]
        rows = np.array([4 * i + α for α in range(4)])
        cols = np.arange(M)
        data = np.repeat(1, M)

        B = CscMatrix((data, (rows, cols)), shape=(N, M))

        # Calculate the density of states.
        ρ = {}
        for ε_n, Γ_n in zip(ε, Γ):
            # Solve the linear equations for the resolvent.
            A = (ε_n + 1j * Γ_n) * I - H
            X = spla.spsolve(A, B)

            # Extract the few elements of interest.
            x = X.multiply(B).sum(axis=0)

            # Calculate and store the density of states.
            e_up = x[0, 0]
            e_dn = x[0, 1]
            h_up = x[0, 2]
            h_dn = x[0, 3]

            ρ[+ε_n] = -np.imag(e_up + e_dn) / π
            ρ[-ε_n] = -np.imag(h_up + h_dn) / π

        # Extract only the requested values.
        ρ = np.array([ρ[ε_n] for ε_n in energies])

        return ρ


def swave() -> Callable:
    """Hamiltonian terms for s-wave superconducting order.

    If you let `σ_s = swave()`, then the on-site pairing term for an
    s-wave superconductor is given by `Δ[i, i] = Δ_s(i) * σ_s(i)`.
    Similarly, extended s-wave order can be obtained by setting
    `Δ[i, j] = Δ_s(i, j) * σ_s(i, j) for nearest-neighbor sites
    `i` and `j`. The prefactors `Δ_s` above can be complex scalar
    functions of the lattice site indices, but can often be set
    constant in non-selfconsistent calculations without currents.
    """

    def σ_s(*_):
        return jσ2

    return σ_s


def pwave(dvector: str) -> Callable:
    """Hamiltonian terms for p-wave superconducting order.

    When calling this function, you must provide a d-vector expression
    as a string [e.g. `σ_p = pwave("(p_x + jp_y) * (e_x + je_y)")`] to
    construct the a p-wave triplet superconducting order parameter.
    The result is a function `σ_p(i, j)` that depend on two
    nearest-neighbor lattice sites `i` and `j`. The superconducting
    pairing is then given by `Δ[i, j] = Δ_p(i, j) * σ_p(i, j)`, where
    `Δ_p` can be any complex scalar function of position `(i + j)/2`.
    It is often sufficient to set the prefactor `Δ_p` to a constant.

    The algorithm implemented here is explained in Sec. II-B in:

    Ouassou et al. PRB 109, 174506 (2024).
    DOI: 10.1103/PhysRevB.109.174506
    """
    # Basis vectors for spin axes.
    e_x = np.array([[1], [0], [0]])
    e_y = np.array([[0], [1], [0]])
    e_z = np.array([[0], [0], [1]])

    je_x = 1j * e_x
    je_y = 1j * e_y
    je_z = 1j * e_z

    # Basis vectors for momentum.
    p_x = e_x.T
    p_y = e_y.T
    p_z = e_z.T

    jp_x = 1j * p_x
    jp_y = 1j * p_y
    jp_z = 1j * p_z

    # Convert the d-vector expression to a 3x3 numerical matrix.
    D = eval(dvector)

    # Construct gap matrix Δ(p) = [d(p)⋅σ] jσ2 = [(D'p) ⋅ σ] jσ2.
    # In practice, we do this by calculating Δ = D'σ jσ2, such
    # that we simply end up with the gap matrix Δ(p) = Δ ⋅ p.
    Δ = np.einsum("kp,kab,bc -> pac", D, σ, jσ2) / 2

    # Function for evaluating Δ(p) on the lattice.
    def σ_p(i: Coord, j: Coord) -> Matrix:
        δ = np.subtract(j, i)
        return np.einsum("iab,i -> ab", Δ, δ)

    return σ_p


def dwave() -> Callable:
    """Generate the d-wave superconducting order parameter.

    This function returns a function `σ_d(i, j)` that takes two
    nearest-neighbor lattice sites `i` and `j` as arguments. The
    d-wave pairing is given by `Δ[i, j] = Δ_d(i, j) * σ_d(i, j)`,
    where the prefactor `Δ_d` can be any complex scalar function
    of position `(i+j)/2`. Often, it can be set to a constant.

    We specifically consider the d_{x^2 - y^2} order parameter on a
    presumably square lattice. This means that the order parameter
    should have a momentum structure ~ (p_x^2 - p_y^2)/p_F^2 and spin
    structure ~ jσ_2 (spin-singlet). It might work on some non-square
    lattices as well, but this has not been checked by the author.
    """

    def σ_d(i: Coord, j: Coord) -> Matrix:
        δ = np.subtract(j, i)
        Δ_ij = (δ[0] ** 2 - δ[1] ** 2) / (np.sum(δ**2) + 1e-16)

        return Δ_ij * jσ2

    return σ_d


def ssd(system: Hamiltonian) -> Callable:
    """Sine-Squared Deformation of a Hamiltonian on a cubic lattice.

    When you call this function as `φ = ssd(system)`, where `system`
    is an instance of the `Hamiltonian` class, you obtain a function
    `φ(i, j)` that depends on two lattice coordinates `i` and `j`. The
    usage of this function is to include `φ(i, j)` as a prefactor for
    every off-site term `H[i, j]` and `Δ[i, j]` in the Hamiltonian,
    and add `φ(i, i)` as a prefactor for every on-site term `H[i, i]`
    and `Δ[i, i]`. This approach is known to reduce finite-size
    effects in real-space simulations, which can be especially
    important in the case of incommensurate order on the lattice.

    For more information about how this approach can be useful, see
    the following reference as well as the references therein:

    Hodt et al. PRB 107, 224427 (2023).
    DOI: 10.1103/PhysRevB.107.224427

    TODO: Consider integrating this into `Hamiltonian.__exit__` to
    make it more transparent to the user (if it proves very useful).
    """

    # Define the profile φ(i, j) used in the SSD method.
    def profile(i: Coord, j: Coord):
        # Determine the origin and maximum radius for the system. Since we use
        # corner-centered coordinates, these two values are actually the same.
        # Note however the offset to get a maximum coordinate from a shape.
        R = np.array([(N - 1) for N in system.lattice.shape]) / 2

        # Determine the distance vectors of the provided coordinates
        # i = (i_x, i_y, i_z), j = (j_x, j_y, j_z) from the origin.
        r_i = np.array(i) - R
        r_j = np.array(j) - R
        r = (r_i + r_j) / 2

        # We now consider only the magnitude of the distances.
        R = la.norm(R)
        r = la.norm(r)

        # Calculate the sine-squared deformation.
        return (1 / 2) * (1 + np.cos(π * r / (R + 1 / 2)))

    return profile



import numpy as np
from pytest import raises
from scipy.linalg import eigh

from bodge import *
from bodge.common import *


def test_hermitian():
    # Instantiate a somewhat dense complex Hamiltonian. Note that
    # the submatrices need to be Hermitian for the whole to be so.
    lattice = CubicLattice((3, 5, 7))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = 1 * σ3 + 2 * σ2
            Δ[i, i] = 5 * σ0 - 3 * σ2

        for i, j in lattice.bonds():
            H[i, j] = 3 * σ0 - 4 * σ2
            Δ[i, j] = 2 * σ3 + 5 * σ2

        for i, j in lattice.edges():
            H[i, j] = 7 * σ0 - 13 * σ2
            Δ[i, j] = 9 * σ3 + 11 * σ2

    # Verify that the result is Hermitian.
    H = system._matrix.todense()
    assert np.allclose(H, H.T.conj())

    # Check non-hermiticity warnings.
    with raises(RuntimeError):
        i = (1, 1, 1)
        with system as (H, Δ):
            H[i, i] = 1j * σ1


def test_compilation():
    # Instantiate a somewhat random test system.
    lattice = CubicLattice((3, 5, 7))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i, j in lattice:
            H[i, j] = 3 * σ0 - 4 * σ2
            Δ[i, j] = 2 * σ3 + 5 * σ2

    # Construct matrix instances in different formats.
    H_BSR = system.matrix(format="bsr")
    H_CSR = system.matrix(format="csr")
    H_CSC = system.matrix(format="csc")
    H_DNS = system.matrix(format="dense")

    # Check that the dense matrix looks right.
    assert np.real(H_DNS[0, 0]) == 3
    assert np.imag(H_DNS[0, 1]) == 4
    assert np.real(H_DNS[0, 2]) == 2
    assert np.imag(H_DNS[0, 3]) == -5

    # Test that each construction is equivalent.
    assert np.max(np.abs(H_BSR - H_DNS)) < 1e-6
    assert np.max(np.abs(H_CSR - H_DNS)) < 1e-6
    assert np.max(np.abs(H_CSC - H_DNS)) < 1e-6

    # Check that the BSR blocksize is correct.
    assert H_BSR.blocksize == (4, 4)

    # Test error handling.
    with raises(Exception):
        H = system.matrix(format="blah")
    with raises(Exception):
        H = system.matrix(format=1)


def test_swave_hermitian():
    """Test that s-wave superconducting Hamiltonians are Hermitian."""
    lattice = CubicLattice((10, 10, 1))
    system = Hamiltonian(lattice)

    σ_s = swave()
    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = -0.1 * σ0
            Δ[i, i] = -0.1j * σ_s(i, i)  # On-site s-wave

        for i, j in lattice.bonds():
            H[i, j] = -1 * σ0
            Δ[i, j] = -0.3 * σ_s(i, j)  # Extended s-wave

    H = system._matrix.todense()
    assert np.allclose(H, H.T.conj())


def test_pwave_basic():
    """Brute-force test all d(p) = e_i p_j cases."""
    Δ = pwave("e_x * p_x")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), σ1 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_x * p_y")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), σ1 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_x * p_z")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), σ1 @ jσ2 / 2)

    Δ = pwave("e_y * p_x")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), σ2 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_y * p_y")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), σ2 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_y * p_z")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), σ2 @ jσ2 / 2)

    Δ = pwave("e_z * p_x")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), σ3 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_z * p_y")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), σ3 @ jσ2 / 2)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), 0)

    Δ = pwave("e_z * p_z")
    assert np.allclose(Δ((0, 0, 0), (1, 0, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 1, 0)), 0)
    assert np.allclose(Δ((0, 0, 0), (0, 0, 1)), σ3 @ jσ2 / 2)


def test_pwave_symmetries():
    """Check the p-wave property Δ(i, j) = -Δ(j, i)."""
    for desc in [
        "e_x * p_x",
        "e_z * p_y",
        "e_y * jp_z",
        "e_z * (p_x + jp_y)",
        "(e_x + je_y) * (p_y + jp_z)",
    ]:
        Δ = pwave(desc)
        for x in range(3):
            for y in range(3):
                for z in range(3):
                    i0 = (x, y, z)
                    j1 = (x + 1, y, z)
                    j2 = (x, y + 1, z)
                    j3 = (x, y, z + 1)

                    assert np.allclose(+Δ(i0, j1), -Δ(j1, i0))
                    assert np.allclose(+Δ(i0, j2), -Δ(j2, i0))
                    assert np.allclose(+Δ(i0, j3), -Δ(j3, i0))


def test_pwave_hermitian():
    """Test that d-vector-constructed Hamiltonians are Hermitian."""
    lattice = CubicLattice((10, 10, 1))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = -0.1 * σ0

        for i, j in lattice.bonds():
            H[i, j] = -1 * σ0

    for desc in [
        "e_x * p_x",
        "e_z * p_y",
        "e_y * jp_z",
        "e_z * (p_x + jp_y)",
        "(e_x + je_y) * (p_y + jp_z)",
    ]:
        Δ_p = pwave(desc)
        with system as (H, Δ):
            for i, j in lattice.bonds():
                Δ[i, j] = -0.1 * Δ_p(i, j)

        H = system._matrix.todense()
        assert np.allclose(H, H.T.conj())


def test_dwave_symmetries():
    """Test that the d_{x^2 - y^2} order parameter behaves as expected."""
    Δ_d = dwave()

    # Zero on-site contributions.
    assert np.allclose(Δ_d((0, 0, 0), (0, 0, 0)), 0 * σ0)
    assert np.allclose(Δ_d((1, 2, 3), (1, 2, 3)), 0 * σ0)

    # Zero z-axis contributions.
    assert np.allclose(Δ_d((0, 0, 0), (0, 0, 1)), 0 * σ0)
    assert np.allclose(Δ_d((0, 0, 1), (0, 0, 0)), 0 * σ0)

    # Positive x-axis contributions.
    assert np.allclose(Δ_d((0, 0, 0), (1, 0, 0)), +1 * jσ2)
    assert np.allclose(Δ_d((0, 0, 0), (9, 0, 0)), +1 * jσ2)
    assert np.allclose(Δ_d((1, 0, 0), (0, 0, 0)), +1 * jσ2)
    assert np.allclose(Δ_d((9, 0, 0), (0, 0, 0)), +1 * jσ2)

    # Negative y-axis contributions.
    assert np.allclose(Δ_d((0, 0, 0), (0, 1, 0)), -1 * jσ2)
    assert np.allclose(Δ_d((0, 0, 0), (0, 9, 0)), -1 * jσ2)
    assert np.allclose(Δ_d((0, 1, 0), (0, 0, 0)), -1 * jσ2)
    assert np.allclose(Δ_d((0, 9, 0), (0, 0, 0)), -1 * jσ2)

    # Zero diagonal contributions.
    assert np.allclose(Δ_d((1, +1, 0), (0, 0, 0)), 0 * σ0)
    assert np.allclose(Δ_d((1, -1, 0), (0, 0, 0)), 0 * σ0)
    assert np.allclose(Δ_d((0, 0, 0), (1, +1, 0)), 0 * σ0)
    assert np.allclose(Δ_d((0, 0, 0), (1, -1, 0)), 0 * σ0)


def test_dwave_hermitian():
    """Test that d-wave superconductors have Hermitian Hamiltonians."""
    lattice = CubicLattice((10, 10, 1))
    system = Hamiltonian(lattice)
    Δ_d = dwave()

    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = -0.1 * σ0

        for i, j in lattice.bonds():
            H[i, j] = -1 * σ0
            Δ[i, j] = -0.1 * Δ_d(i, j)

    H = system._matrix.todense()
    assert np.allclose(H, H.T.conj())


def test_ssd():
    """Test the mathematical properties of the SSD profile."""
    lattice = CubicLattice((31, 137, 1))
    system = Hamiltonian(lattice)
    φ = ssd(system)

    # The profile goes to zero at corners.
    i = (0, 0, 0)
    assert np.allclose(φ(i, i), 0, atol=0.001)

    # The profile goes to one at origin.
    i = (15, 68, 0)
    assert np.allclose(φ(i, i), 1, atol=0.001)

    # Test the symmetry of the solution.
    # Useful to catch off-by-one errors.
    i = (0, 0, 0)
    j = (30, 136, 0)
    assert φ(i, i) == φ(j, j)

    # Test that coordinate averaging works.
    # The result here should be exact.
    i = (1, 21, 0)
    j = (11, 1, 0)
    k = (6, 11, 0)
    assert φ(i, j) == φ(k, k)


def test_diagonalize():
    # Instantiate a system with superconductivity and a barrier.
    lattice = CubicLattice((10, 10, 1))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = 4 * σ0
            if i[0] > 5:
                Δ[i, i] = 1 * jσ2
            elif i[0] > 3:
                H[i, i] += 6 * σ0
        for i, j in lattice.bonds():
            H[i, j] = -1 * σ0

    # Calculate the eigenvalues the manual way.
    H = system.matrix(format="dense")
    E, X = eigh(H, subset_by_value=(0, np.inf))
    X = X.T

    # Confirm that we got positive eigenvalues and that we have
    # interpreted the corresponding eigenvector matrix correctly.
    assert E.size == 200
    for n, E_n in enumerate(E):
        assert E_n > 0
        assert np.allclose(H @ X[n, :], E_n * X[n, :])

    # Calculate the same eigenvalues via the package, and ensure
    # that the eigenvalues and eigenvectors are consistent.
    eigval, eigvec = system.diagonalize()
    assert np.allclose(eigval, E)
    for n, E_n in enumerate(E):
        for m in range(100):
            assert np.allclose(eigvec[n, m, 0], X[n, 4 * m + 0])
            assert np.allclose(eigvec[n, m, 1], X[n, 4 * m + 1])
            assert np.allclose(eigvec[n, m, 2], X[n, 4 * m + 2])
            assert np.allclose(eigvec[n, m, 3], X[n, 4 * m + 3])


def test_free_energy():
    # Instantiate a simple S/N/F system.
    lattice = CubicLattice((10, 7, 3))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i in lattice.sites():
            if i[0] <= 3:
                H[i, i] = -0.5 * σ0
                Δ[i, i] = -1.0 * jσ2
            if i[0] >= 7:
                H[i, i] = +0.5 * σ0 + 1.5 * σ3

        for i, j in lattice.bonds():
            H[i, j] = -1 * σ0

    # Verify the expression for free energy.
    ε, χ = system.diagonalize()
    ε = np.hstack([-ε, +ε])
    for T in [0.01, 0.1, 1.0]:
        # Use the predefined function.
        F1 = system.free_energy(T)

        # Use standard expression.
        F2 = -(T / 2) * np.sum(np.log(1 + np.exp(-ε / T)))

        assert np.allclose(F1, F2)

    # Check the zero-temperature result.
    F1 = system.free_energy(0.0)
    F2 = (1 / 2) * np.sum(ε[ε < 0])
    assert np.allclose(F1, F2)


def test_ldos():
    # Instantiate a normal metal.
    lattice = CubicLattice((16, 16, 1))
    system = Hamiltonian(lattice)

    with system as (H, Δ):
        for i in lattice.sites():
            H[i, i] = -1.5 * σ0
        for i, j in lattice.bonds():
            H[i, j] = -1.0 * σ0

    # Calculate the central LDOS.
    Δs = 0.5
    i = (8, 8, 0)
    ω = np.array([-1.2 * Δs, -0.8 * Δs, +0.8 * Δs, 1.2 * Δs])
    ρ1 = system.ldos(i, ω)

    # Add superconductivity.
    with system as (H, Δ):
        for i in lattice.sites():
            Δ[i, i] = Δs * jσ2
    ρ2 = system.ldos(i, ω)

    # The LDOS should decrease inside the gap of a superconductor.
    assert ρ2[1] < ρ1[1]
    assert ρ2[2] < ρ1[2]

    # The LDOS should increase outside the gap of a superconductor.
    assert ρ2[0] > ρ1[0]
    assert ρ2[3] > ρ1[3]



"""Common imports and definitions used throughout the project."""

# Common imports.
import numpy as np
import numpy.typing as npt
import scipy.linalg as la
import scipy.sparse as sp
import scipy.sparse.linalg as spla
from beartype import beartype as typecheck
from beartype.typing import Callable, Iterator

# Data types for working with Lattice coordinates.
Index = int
Coord = tuple[int, int, int]
Indices = tuple[Index, Index]
Coords = tuple[Coord, Coord]

# Data types for working with various matrix formats.
Matrix = npt.NDArray[np.float64] | npt.NDArray[np.complex128]
CooMatrix = sp.coo_matrix
DiaMatrix = sp.dia_matrix
BsrMatrix = sp.bsr_matrix
CsrMatrix = sp.csr_matrix
CscMatrix = sp.csc_matrix
SpMatrix = sp.spmatrix

# Fundamental constants.
π = np.pi

# Pauli matrices used to represent spin.
σ0: Matrix = np.array([[+1, 0], [0, +1]], dtype=np.complex128)
σ1: Matrix = np.array([[0, +1], [+1, 0]], dtype=np.complex128)
σ2: Matrix = np.array([[0, -1j], [+1j, 0]], dtype=np.complex128)
σ3: Matrix = np.array([[+1, 0], [0, -1]], dtype=np.complex128)

σ = np.stack([σ1, σ2, σ3])

# Compact notation for imaginary versions.
jσ0: Matrix = 1j * σ0
jσ1: Matrix = 1j * σ1
jσ2: Matrix = 1j * σ2
jσ3: Matrix = 1j * σ3

jσ = np.stack([jσ1, jσ2, jσ3])

# ASCII alternatives.
pi = π

sigma0 = σ0
sigma1 = σ1
sigma2 = σ2
sigma3 = σ3

sigma = σ

jsigma0 = jσ0
jsigma1 = jσ1
jsigma2 = jσ2
jsigma3 = jσ3

jsigma = jσ
